import { __awaiter } from '../../node_modules/tslib/tslib.es6.mjs.js';
import { fetcher, addParamsToSearch } from '../../client/client.js';
import { servers } from '../../client/openapi/fhir-api.js';
import { getConfig } from '../../config.js';
import { ZapEHRSdkError, ZapEHRFHIRError } from '../../errors/index.js';

function getServer() {
    var _a;
    return (_a = getConfig().ZAPEHR_FHIR_API_URL) !== null && _a !== void 0 ? _a : servers[0];
}
function fhirFetcher(path, method) {
    return (params, request) => __awaiter(this, void 0, void 0, function* () {
        try {
            // must await here to catch
            return yield fetcher(getServer, path, method)(params, request);
        }
        catch (err) {
            // FHIR API error messages are JSON strings
            const fullError = err;
            if (typeof fullError.message === 'string') {
                throw new ZapEHRSdkError({
                    message: fullError.message,
                    code: fullError.code,
                });
            }
            throw new ZapEHRFHIRError({
                error: fullError.message,
                code: fullError.code,
            });
        }
    });
}
/**
 * @deprecated Use `search` instead
 *
 * Performs a FHIR search and returns the results as an array of FHIR resources
 *
 * @param options FHIR resource type and FHIR search parameters
 * @param request optional ZapEHRClientRequest object
 * @returns Array of FHIR resources
 */
function list({ resourceType, params }, request) {
    return __awaiter(this, void 0, void 0, function* () {
        const listResp = yield listAsBundle({ resourceType, params }, request);
        const resources = listResp.entry ? listResp.entry.map((entry) => entry.resource) : [];
        return resources;
    });
}
/**
 * @deprecated Use `search` instead
 *
 * Performs a FHIR search and returns the results as a Bundle resource
 *
 * @param options FHIR resource type and FHIR search parameters
 * @param request optional ZapEHRClientRequest object
 * @returns FHIR Bundle resource
 */
function listAsBundle({ resourceType, params }, request) {
    return __awaiter(this, void 0, void 0, function* () {
        return search({ resourceType, params }, request);
    });
}
/**
 * Performs a FHIR search and returns the results as a Bundle resource
 *
 * @param options FHIR resource type and FHIR search parameters
 * @param request optional ZapEHRClientRequest object
 * @returns FHIR Bundle resource
 */
function search({ resourceType, params }, request) {
    return __awaiter(this, void 0, void 0, function* () {
        let paramMap = {};
        if (params) {
            paramMap = Object.entries(params).reduce((acc, [_, param]) => {
                if (!acc[param.name]) {
                    acc[param.name] = [];
                }
                acc[param.name].push(param.value);
                return acc;
            }, {});
        }
        return fhirFetcher(`/${resourceType}/_search`, 'POST')(paramMap, Object.assign(Object.assign({}, request), { contentType: 'application/x-www-form-urlencoded' }));
    });
}
function create(params, request) {
    return __awaiter(this, void 0, void 0, function* () {
        const { resourceType } = params;
        return fhirFetcher(`/${resourceType}`, 'POST')(params, request);
    });
}
function get({ resourceType, id }, request) {
    return __awaiter(this, void 0, void 0, function* () {
        return fhirFetcher(`/${resourceType}/${id}`, 'GET')({}, request);
    });
}
function update(params, request) {
    return __awaiter(this, void 0, void 0, function* () {
        const { id, resourceType } = params;
        return fhirFetcher(`/${resourceType}/${id}`, 'PUT')(params, Object.assign(Object.assign({}, request), { ifMatch: (request === null || request === void 0 ? void 0 : request.optimisticLockingVersionId) ? `W/"${request.optimisticLockingVersionId}"` : undefined }));
    });
}
function patch({ resourceType, id, operations }, request) {
    return __awaiter(this, void 0, void 0, function* () {
        return fhirFetcher(`/${resourceType}/${id}`, 'PATCH')(operations, Object.assign(Object.assign({}, request), { contentType: 'application/json-patch+json', ifMatch: (request === null || request === void 0 ? void 0 : request.optimisticLockingVersionId) ? `W/"${request.optimisticLockingVersionId}"` : undefined }));
    });
}
function del({ resourceType, id }, request) {
    return __awaiter(this, void 0, void 0, function* () {
        return fhirFetcher(`/${resourceType}/${id}`, 'DELETE')({}, request);
    });
}
function history({ resourceType, id, versionId }, request) {
    return __awaiter(this, void 0, void 0, function* () {
        return fhirFetcher(`/${resourceType}/${id}/_history${versionId ? `/${versionId}` : ''}`, 'GET')({}, request);
    });
}
function batchInputRequestToBundleEntryItem(request) {
    const { method, url } = request;
    const baseRequest = {
        request: {
            method,
            url,
        },
    };
    if (url.split('?').length > 1) {
        const [resource, query] = url.split('?');
        const params = query
            .split('&')
            .map((param) => {
            const [name, value] = param.split('=');
            return { name, value };
        })
            .reduce((acc, { name, value }) => {
            acc[name] = value;
            return acc;
        }, {});
        const search = new URLSearchParams();
        addParamsToSearch(params, search);
        baseRequest.request.url = `${resource}?${search.toString()}`;
    }
    if (['GET', 'DELETE', 'HEAD'].includes(method)) {
        return baseRequest;
    }
    if (method === 'PUT' || method === 'PATCH') {
        const { resource } = request;
        return Object.assign(Object.assign({}, baseRequest), { resource });
    }
    if (method === 'POST') {
        const { resource, fullUrl } = request;
        return Object.assign(Object.assign({}, baseRequest), { resource,
            fullUrl });
    }
    throw new Error('Unrecognized method');
}
function bundleRequest(type) {
    return function (input, request) {
        return __awaiter(this, void 0, void 0, function* () {
            return fhirFetcher('/', 'POST')({
                resourceType: 'Bundle',
                type,
                entry: input.requests.map(batchInputRequestToBundleEntryItem),
            }, request);
        });
    };
}
const batch = bundleRequest('batch');
const transaction = bundleRequest('transaction');

export { batch, create, del as delete, get, history, list, listAsBundle, patch, search, transaction, update };
//# sourceMappingURL=index.js.map
